/* Fo
 * fo-property-util.c: Utility functions for properties.
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-property-util.c,v 1.4 2004/05/02 22:56:37 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include <string.h>
#include "fo-property-util.h"

/**
 * fo_property_util_get_color_initial:
 * 
 * Get the initial color value for borders.
 * 
 * Return value: #FoDatatype for initial border color.
 **/
FoDatatype*
fo_property_util_get_color_initial (void)
{
  return fo_color_get_color_black ();
}

/**
 * fo_property_util_resolve_no_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_util_resolve_no_enum (const gchar *token,
				  FoContext   *context,
				  GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  g_set_error (error,
	       FO_FO_ERROR,
	       FO_FO_ERROR_ENUMERATION_TOKEN,
	       _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
	       "property",
	       token);

  return NULL;
}

/**
 * fo_property_util_resolve_color_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_util_resolve_color_enum (const gchar *token,
				     FoContext   *context,
				     GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "black") == 0)
    {
      return g_object_ref (fo_color_get_color_black ());
    }
  else if (strcmp (token, "silver") == 0)
    {
      return g_object_ref (fo_color_get_color_silver ());
    }
  else if (strcmp (token, "gray") == 0)
    {
      return g_object_ref (fo_color_get_color_gray ());
    }
  else if (strcmp (token, "white") == 0)
    {
      return g_object_ref (fo_color_get_color_white ());
    }
  else if (strcmp (token, "maroon") == 0)
    {
      return g_object_ref (fo_color_get_color_maroon ());
    }
  else if (strcmp (token, "red") == 0)
    {
      return g_object_ref (fo_color_get_color_red ());
    }
  else if (strcmp (token, "purple") == 0)
    {
      return g_object_ref (fo_color_get_color_purple ());
    }
  else if (strcmp (token, "fuchsia") == 0)
    {
      return g_object_ref (fo_color_get_color_fuchsia ());
    }
  else if (strcmp (token, "green") == 0)
    {
      return g_object_ref (fo_color_get_color_green ());
    }
  else if (strcmp (token, "lime") == 0)
    {
      return g_object_ref (fo_color_get_color_lime ());
    }
  else if (strcmp (token, "olive") == 0)
    {
      return g_object_ref (fo_color_get_color_olive ());
    }
  else if (strcmp (token, "yellow") == 0)
    {
      return g_object_ref (fo_color_get_color_yellow ());
    }
  else if (strcmp (token, "navy") == 0)
    {
      return g_object_ref (fo_color_get_color_navy ());
    }
  else if (strcmp (token, "blue") == 0)
    {
      return g_object_ref (fo_color_get_color_blue ());
    }
  else if (strcmp (token, "teal") == 0)
    {
      return g_object_ref (fo_color_get_color_teal ());
    }
  else if (strcmp (token, "aqua") == 0)
    {
      return g_object_ref (fo_color_get_color_aqua ());
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "color property",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_resolve_auto_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_util_resolve_auto_enum (const gchar *token,
				    FoContext   *context,
				    GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "auto") == 0)
    {
      return g_object_ref (fo_enum_get_enum_auto ());
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "property",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_resolve_auto_always_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_util_resolve_auto_always_enum (const gchar *token,
					   FoContext   *context,
					   GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "auto") == 0)
    {
      return g_object_ref (fo_enum_get_enum_auto ());
    }
  else if (strcmp (token, "always") == 0)
    {
      return g_object_ref (fo_enum_get_always ());
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "property",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_resolve_force_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_util_resolve_force_enum (const gchar *token,
				     FoContext   *context,
				     GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "force") == 0)
    {
      return g_object_ref (fo_enum_get_enum_by_nick (token));
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "property",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_validate_color:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property
 * @context:  #FoContext object from which to possibly inherit values
 * @error:    Information about any error that has occurred
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed
 * 
 * Return value: Valid datatype value or NULL
 **/
FoDatatype*
fo_property_util_validate_color (FoDatatype *datatype,
				 FoContext  *context,
				 GError    **error)
{
  FoDatatype *new_datatype;
  GError     *tmp_error = NULL;
  gchar      *token;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_COLOR (datatype))
    {
      return datatype;
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
	fo_property_util_resolve_color_enum (token, context, NULL);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
	fo_property_util_resolve_color_enum (token, context, NULL);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   "color property",
		   fo_object_sprintf (datatype),
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      return NULL;
    }
}

/**
 * fo_property_util_resolve_condity_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_util_resolve_condity_enum (const gchar *token,
				       FoContext   *context,
				       GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "discard") == 0)
    {
      return g_object_ref (fo_boolean_get_false ());
    }
  else if (strcmp (token, "retain") == 0)
    {
      return g_object_ref (fo_boolean_get_true ());
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "conditionality",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_resolve_ipdim_percent:
 * @percentage: Percentage value to resolve.
 * @font_size:  Font size to use if a percentage of font-size.
 * @fo_node:    Current #FoFo.
 * @context:    #FoContext of parent #FoFo.
 * @error:      Indicates whether an error occurs.
 * 
 * Resolves @percentage as a percentage of a known value which,
 * depending on the property, could be the current font-size value or
 * some property value of the current #FoFo or its first #FoArea.
 * 
 * Return value: #FoDatatype representing resolved value.
 */
FoDatatype *
fo_property_util_resolve_ipdim_percent (gdouble           percentage,
					const FoDatatype *font_size,
					const FoFo       *fo_node,
					const FoContext  *context,
					GError          **error)
{
  FoFoAreaIterator *iterator;
  FoArea *fo_area;
  gdouble width;

  g_return_val_if_fail (FO_IS_FO (fo_node), NULL);

  /* FIXME: Need to get content-rectangle of closest ancestor
     block-area that is not a line-area. */
  iterator = fo_fo_get_area_iterator (FO_FO (fo_node_parent (FO_NODE (fo_node))));

  fo_area = fo_fo_area_iterator_get_area (iterator);

  g_free (iterator);

  g_assert (fo_area != NULL);

  width = fo_area_get_child_available_ipdim (fo_area);

  return fo_length_new_with_value (width * percentage * 0.01);
}


/**
 * fo_property_util_get_width_initial:
 * 
 * Get the initial width value for borders.
 * 
 * Return value: #FoDatatype for initial border width.
 **/
FoDatatype*
fo_property_util_get_width_initial (void)
{
  return fo_length_get_width_medium ();
}

/**
 * fo_property_util_resolve_width_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token
 * @context: #FoContext object from which to possibly inherit values
 * @error:   Information about an error that has occurred
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL
 * 
 * Return value: Resolved enumeration value or NULL
 **/
FoDatatype*
fo_property_util_resolve_width_enum (const gchar *token,
				     FoContext   *context,
				     GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "thin") == 0)
    {
      return g_object_ref (fo_length_get_width_thin ());
    }
  else if (strcmp (token, "medium") == 0)
    {
      return g_object_ref (fo_length_get_width_medium ());
    }
  else if (strcmp (token, "thick") == 0)
    {
      return g_object_ref (fo_length_get_width_thick ());
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "border width",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_validate_width:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property.
 * @context:  #FoContext object from which to possibly inherit values.
 * @error:    Information about any error that has occurred.
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed.
 * 
 * Return value: Valid datatype value or NULL.
 **/
FoDatatype*
fo_property_util_validate_width (FoDatatype *datatype,
				 FoContext  *context,
				 GError    **error)
{
  FoDatatype *new_datatype;
  GError     *tmp_error = NULL;
  gchar      *token;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_ENUM (datatype))
    {
      return datatype;
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
        fo_property_util_resolve_width_enum (token, context, &tmp_error);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      g_object_unref (datatype);

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
        fo_property_util_resolve_width_enum (token, context, &tmp_error);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      g_object_unref (datatype);

      return new_datatype;
    }
  else if (FO_IS_LENGTH (datatype))
    {
      return datatype;
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   "border width",
		   fo_object_sprintf (datatype),
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      return NULL;
    }
}

/**
 * fo_property_util_get_style_initial:
 * 
 * Get the initial style value for borders.
 * 
 * Return value: #FoDatatype for initial border style.
 **/
FoDatatype*
fo_property_util_get_style_initial (void)
{
  return fo_enum_get_enum_none ();
}

/**
 * fo_property_util_resolve_style_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token
 * @context: #FoContext object from which to possibly inherit values
 * @error:   Information about an error that has occurred
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL
 * 
 * Return value: Resolved enumeration value or NULL
 **/
FoDatatype*
fo_property_util_resolve_style_enum (const gchar *token,
				     FoContext   *context,
				     GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if ((strcmp (token, "none") == 0) ||
      (strcmp (token, "hidden") == 0) ||
      (strcmp (token, "dotted") == 0) ||
      (strcmp (token, "dashed") == 0) ||
      (strcmp (token, "solid") == 0) ||
      (strcmp (token, "double") == 0) ||
      (strcmp (token, "groove") == 0) ||
      (strcmp (token, "ridge") == 0) ||
      (strcmp (token, "inset") == 0) ||
      (strcmp (token, "outset") == 0))
    {
      return g_object_ref (fo_enum_get_enum_by_nick (token));
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   "border style",
		   token);
      return NULL;
    }
}

/**
 * fo_property_util_validate_style:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property
 * @context:  #FoContext object from which to possibly inherit values
 * @error:    Information about an error that has occurred
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed
 * 
 * Return value: Valid datatype value or NULL
 **/
FoDatatype*
fo_property_util_validate_style (FoDatatype *datatype,
				 FoContext  *context,
				 GError    **error)
{
  FoDatatype *new_datatype;
  GError *tmp_error = NULL;
  gchar *token;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_ENUM (datatype))
    {
      return datatype;
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
	fo_property_util_resolve_style_enum (token, context, &tmp_error);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      g_object_unref (datatype);

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
	fo_property_util_resolve_style_enum (token, context, &tmp_error);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      g_object_unref (datatype);

      return new_datatype;
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   "border style",
		   fo_object_sprintf (datatype),
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      return NULL;
    }
}
